"""
Module that contains function common to both nuclei (HE/Fluo) and mitochondria analysis (SDH)
"""
import os

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
import sys
import tensorflow as tf
import torch
from cellpose.models import Cellpose
from csbdeep.utils import normalize
from stardist.models import StarDist2D
from tensorflow import keras
import numpy as np
from PIL import Image

# from .gradcam import make_gradcam_heatmap, save_and_display_gradcam
from .random_brightness import RandomBrightness

tf.random.set_seed(42)
np.random.seed(42)

if len(tf.config.list_physical_devices("GPU")) >= 1:
    use_GPU = True
    tf.config.experimental.set_memory_growth(
        tf.config.list_physical_devices("GPU")[0], True
    )

else:
    use_GPU = False


class HiddenPrints:
    """
    Class to hide the print of some function during the CLI analysis
    """

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def is_gpu_availiable():
    """
    Check if the GPU is available for the neural network inference
    """
    return use_GPU


def load_cellpose():
    """
    Load and return the cellpose model
    """
    model_c = Cellpose(gpu=use_GPU, model_type="cyto2")
    return model_c


def load_stardist(fluo=False):
    """Load and return the stardist model in the right mode (fluo or HE)

    Args:
        fluo (bool, optional): If true return Fluo StarDist Model. Defaults to False.

    Returns:
        StarDist Model instance: the stardist model
    """
    if fluo:
        model_s = StarDist2D.from_pretrained("2D_versatile_fluo")
    else:
        model_s = StarDist2D.from_pretrained("2D_versatile_he")
    return model_s


def load_sdh_model(model_path: str):
    """Load and return the SDH model

    Args:
        model_path (str): path to the SDH model

    Returns:
        Keras model instance: tensorflow keras model
    """
    model_sdh = keras.models.load_model(
        model_path, custom_objects={"RandomBrightness": RandomBrightness}
    )
    return model_sdh


@torch.no_grad()
def run_cellpose(image, model_cellpose, diameter=None):
    """Run Cellpose analysis on input image

    Args:
        image (nd.array): Numpy array of the image
        model_cellpose (PyTorch Model): Cellpose model object loaded with load_cellpose()
        diameter (int, optional): Cell diameter estimation for cellpose. Defaults to None.

    Returns:
        nd.array: Label map (mask) of the cellpose prediction
    """
    channel = [[0, 0]]
    with torch.no_grad():
        mask_cellpose, _, _, _ = model_cellpose.eval(
            image, diameter=diameter, channels=channel
        )
        return mask_cellpose


def run_stardist(image, model_stardist, nms_thresh=0.4, prob_thresh=0.5):
    """Run StarDist analysis on input image

    Args:
        image (nd.array): Numpy array of the image
        model_stardist (Stardist Model): Stardist model object loaded with load_stardist()
        nms_thresh (float, optional): NMS Threshold for Stardist. Defaults to 0.4.
        prob_thresh (float, optional): Probability Threshold for Stardist. Defaults to 0.5.

    Returns:
        nd.array: Label map (mask) of the stardist prediction
    """
    img_norm = image / 255
    img_norm = normalize(img_norm, 1, 99.8)
    mask_stardist, _ = model_stardist.predict_instances(
        img_norm, nms_thresh=nms_thresh, prob_thresh=prob_thresh
    )
    return mask_stardist


def label2rgb(img_ndarray, label_map):
    """Convert a label map to RGB image

    Args:
        img_ndarray (nd.array): Numpy array of the image
        label_map (nd.array): Numpy array of the label map (mask)

    Returns:
        nd.array: RGB image of the label map with 3 colors (white, green, red) for 3 classes (background, control, abnormal)
    """
    label_to_color = {
        0: [255, 255, 255],
        1: [15, 157, 88],
        2: [219, 68, 55],
    }
    img_rgb = np.zeros((img_ndarray.shape[0], img_ndarray.shape[1], 3), dtype=np.uint8)

    for gray, rgb in label_to_color.items():
        img_rgb[label_map == gray, :] = rgb
    return img_rgb


def blend_image_with_label(image, label_rgb, fluo=False):
    """Blend the image with the label map together

    Args:
        image (nd.array): Numpy array of the image
        label_rgb (nd.array): Numpy array of the label map (mask)
        fluo (bool, optional): Indicate if the image is fluo (single channel). Defaults to False.

    Returns:
        PIL Image: An overlay image of the image and the label map with a 0.5 alpha
    """
    image = Image.fromarray(image)
    label_rgb = Image.fromarray(label_rgb)
    if fluo:
        image = image.convert("RGB")
    return Image.blend(image, label_rgb, 0.5)



"""
Module that contains the main function to run the ATP analysis
"""
import time
from pathlib import Path
import typer
from rich.console import Console
from rich.table import Table
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn

app = typer.Typer()
console = Console()
table = Table(title="Analysis Results 🥳")


@app.command()
def atp_analysis(
    image_path: Path = typer.Argument(
        ...,
        help="The ATP image file path to analyse.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    mask_path: Path = typer.Option(
        None,
        help="The path to a binary mask to hide slide region during analysis. It needs to be of the same resolution as input image and only pixel marked as 1 will be analyzed.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    cellpose_path: Path = typer.Option(
        None,
        help="The pre-computed CellPose mask to use for analysis. Will run Cellpose if no path provided. Required as an image file.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    output_path: Path = typer.Option(
        None,
        help="The path to the folder to save the results. Will save in the same folder as input image if not specified.",
    ),
    intensity_threshold: int = typer.Option(
        None,
        min=1,
        max=254,
        help="Fiber intensity threshold to differenciate between the two fiber types. If not specified, the analysis will try to deduce it.",
    ),
    cellpose_diameter: int = typer.Option(
        None,
        help="Approximative single cell diameter in pixel for CellPose detection. If not specified, Cellpose will try to deduce it.",
    ),
    export_map: bool = typer.Option(
        True,
        help="Export the original image with cells painted by classification label.",
    ),
    export_stats: bool = typer.Option(
        True, help="Export per fiber and per nuclei stat table."
    ),
):
    """Run the fibre type 1 vs type 2 analysis on ATP images.
    First input arguments and option are printed in stdout and all modules are imported. Then the input image is mask with the binary mask if provided.
    Then depending on the presence of cellpose , Cellpose is run or not and mask accordingly if binary mask is provided.
    Finally the ATP analysis is run with run_atp_analysis() function and the results are saved in the output folder and some info are printed in stdout.
    """
    start_time = time.time()
    console.print(
        "👋 [bold dark_orange]Welcome to the fiber type 1 vs 2 analysis (ATP images)",
    )

    # Print input arguments and options
    console.print(f"📄 INPUT: raw image: {image_path}", style="blue")

    if cellpose_path is None:
        console.print(
            "💡 INFO: No CellPose mask provided, will run CellPose during the analysis.",
            style="blue",
        )
    else:
        console.print(f"📄 INPUT: CellPose mask: {cellpose_path}", style="blue")

    if mask_path is not None:
        console.print(f"📄 INPUT: binary mask: {mask_path}", style="blue")

    # Import all modules
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        transient=False,
    ) as progress:
        progress.add_task(description="Importing the libraries...", total=None)
        from ..src.common_func import (
            is_gpu_availiable,
            load_cellpose,
            run_cellpose,
            label2rgb,
            blend_image_with_label,
            HiddenPrints,
        )
        from ..src.ATP_analysis import run_atp_analysis
        import numpy as np
        from PIL import Image

        try:
            from imageio.v2 import imread
        except ImportError:
            from imageio import imread

    if output_path is None:
        output_path = image_path.parents[0]
    else:
        Path(output_path).mkdir(parents=True, exist_ok=True)

    if is_gpu_availiable():
        console.print("💡 INFO: GPU is available.", style="blue")
    else:
        console.print("❌ INFO: GPU is not available. Using CPU only.", style="red")

    # Load raw image, binary mask, cellpose mask if provided.
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        transient=False,
    ) as progress:
        progress.add_task(description="Reading all inputs...", total=None)
        image_ndarray = imread(image_path)

        if mask_path is not None:
            mask_ndarray = imread(mask_path)
            if np.unique(mask_ndarray).shape[0] != 2:
                console.print(
                    "The mask image should be a binary image with only 2 values (0 and 1).",
                    style="red",
                )
                raise ValueError
            if len(image_ndarray.shape) > 2:
                mask_ndarray = np.repeat(
                    mask_ndarray.reshape(
                        mask_ndarray.shape[0], mask_ndarray.shape[1], 1
                    ),
                    image_ndarray.shape[2],
                    axis=2,
                )
            image_ndarray = image_ndarray * mask_ndarray
        if cellpose_path is not None:
            mask_cellpose = imread(cellpose_path)

    # Run Cellpose if no mask provided
    if cellpose_path is None:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            transient=False,
        ) as progress:
            progress.add_task(description="Running CellPose...", total=None)
            model_cellpose = load_cellpose()
            mask_cellpose = run_cellpose(
                image_ndarray, model_cellpose, cellpose_diameter
            )
            mask_cellpose = mask_cellpose.astype(np.uint16)
            cellpose_mask_filename = image_path.stem + "_cellpose_mask.tiff"
            Image.fromarray(mask_cellpose).save(output_path / cellpose_mask_filename)
        console.print(
            f"💾 OUTPUT: CellPose mask saved as {output_path/cellpose_mask_filename}",
            style="green",
        )

    # If binary mask provided, mask cellpose and stardist mask
    if mask_path is not None:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            transient=False,
        ) as progress:
            progress.add_task(
                description="Masking Cellpose mask with binary mask...",
                total=None,
            )
            mask_ndarray = imread(mask_path)
            mask_cellpose = mask_cellpose * mask_ndarray

    # Run the fiber type 1 vs 2 analysis and get the results table, label map and dataframes
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        transient=False,
    ) as progress:
        progress.add_task(description="Detecting fiber types...", total=None)
        result_df, full_label_map, df_cellpose_details = run_atp_analysis(
            image_ndarray, mask_cellpose, intensity_threshold
        )
    if export_map:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            transient=False,
        ) as progress:
            progress.add_task(
                description="Blending label and original image together...", total=None
            )
            labelRGB_map = label2rgb(image_ndarray, full_label_map)
            overlay_img = blend_image_with_label(image_ndarray, labelRGB_map)
            overlay_filename = image_path.stem + "_label_blend.tiff"
            overlay_img.save(output_path / overlay_filename)

    # Construct the summary table, print all output in stdout and save files in output folder.
    table.add_column("Feature", justify="left", style="cyan")
    table.add_column("Raw Count", justify="center", style="magenta")
    table.add_column("Proportion (%)", justify="right", style="green")
    for index, row in result_df.iterrows():
        table.add_row(
            str(row[0]),
            str(row[1]),
            str(row[2]),
        )
    console.print(table)
    csv_name = image_path.stem + "_results_summary.csv"
    cell_details_name = image_path.stem + "_cell_details.csv"
    result_df.to_csv(
        output_path / csv_name,
        index=False,
    )
    console.print(
        f"💾 OUTPUT: Summary Table saved as {output_path/csv_name}",
        style="green",
    )
    if export_map:
        console.print(
            f"💾 OUTPUT: Overlay image saved as {output_path/overlay_filename}",
            style="green",
        )
    if export_stats:
        df_cellpose_details.drop("image", axis=1).to_csv(
            output_path / cell_details_name,
            index=False,
        )
        console.print(
            f"💾 OUTPUT: Cell Table saved as {output_path/cell_details_name}",
            style="green",
        )
    label_map_name = image_path.stem + "_label_map.tiff"
    Image.fromarray(full_label_map).save(output_path / label_map_name)
    console.print(
        f"💾 OUTPUT: Label map saved as {output_path/label_map_name}", style="green"
    )
    console.print("--- %s seconds ---" % (int(time.time() - start_time)))



"""
Module that contains function used in the nuclei analysis to draw line and calculate coordinates
"""
import math


def line_equation(x1, y1, x2, y2):
    """Calculate the staight line equation given two points

    Args:
        x1 (float): x1 coordinate
        y1 (float): y1 coordinate
        x2 (float): x2 coordinate
        y2 (float): y2 coordinate

    Raises:
        ValueError: If the line equation contains infinite slope or intercept

    Returns:
        m, b (float, float): slope and intercept of the straight line
    """
    # Poor man's handling of the case where the equation is infinite
    if x2 - x1 == 0 and y2 > y1:
        m = 9999
    elif x2 - x1 == 0 and y1 > y2:
        m = -9999
    else:
        m = (y2 - y1) / (x2 - x1)
    b = y1 - m * x1
    if math.isinf(m) or math.isinf(b):
        raise ValueError("Line equation is infinite")
    return m, b


def calculate_intersection(m, b, image_dim=(256, 256)):
    """Calculate the intersection of a line with the image boundaries

    Args:
        m (float): slope of the line
        b (float): intercept of the line
        image_dim (tuple, optional): Image dimension. Defaults to (256, 256).

    Raises:
        ValueError: If no intersection is found

    Returns:
        list of 2 tuples: the two intersection points coordinates
    """
    x = [0, image_dim[1]]
    y = [0, image_dim[0]]
    results = []
    for i in x:
        intersect = m * i + b
        if intersect >= 0 and intersect < image_dim[0]:
            results.append((i, intersect))
    for i in y:
        intersect = (i - b) / m
        if intersect >= 0 and intersect < image_dim[1]:
            results.append((intersect, i))
    if results == []:
        raise ValueError("No intersection found")
    return results


def calculate_distance(x1, y1, x2, y2):
    """Calculate the distance between two points

    Args:
        x1 (float): x1 coordinate
        y1 (float): y1 coordinate
        x2 (float): x2 coordinate
        y2 (float): y2 coordinate

    Returns:
        float: distance between the two points
    """
    return ((x1 - x2) ** 2 + (y1 - y2) ** 2) ** 0.5


def calculate_closest_point(x1, y1, intersections):
    """Calculate the closest intersection point to the first point

    Args:
        x1 (float): x1 coordinate
        y1 (flaot): y1 coordinate
        intersections (list of tuples): list of intersection points

    Returns:
        tuple: closest intersection point coordinates
    """
    distances = []
    for coords in intersections:
        distances.append(calculate_distance(x1, y1, coords[0], coords[1]))
    return intersections[distances.index(min(distances))]



import numpy as np
import pandas as pd
from skimage.draw import line
from skimage.measure import regionprops_table

from .draw_line import *


def extract_ROIs(histo_img, index, cellpose_df, mask_stardist):
    single_cell_img = histo_img[
        cellpose_df.iloc[index, 5] : cellpose_df.iloc[index, 7],
        cellpose_df.iloc[index, 6] : cellpose_df.iloc[index, 8],
    ].copy()
    nucleus_single_cell_img = mask_stardist[
        cellpose_df.iloc[index, 5] : cellpose_df.iloc[index, 7],
        cellpose_df.iloc[index, 6] : cellpose_df.iloc[index, 8],
    ].copy()
    single_cell_mask = cellpose_df.iloc[index, 9]
    single_cell_img[~single_cell_mask] = 0
    nucleus_single_cell_img[~single_cell_mask] = 0

    props_nuc_single = regionprops_table(
        nucleus_single_cell_img,
        intensity_image=single_cell_img,
        properties=[
            "label",
            "area",
            "centroid",
            "eccentricity",
            "bbox",
            "image",
            "perimeter",
            "feret_diameter_max",
        ],
    )
    df_nuc_single = pd.DataFrame(props_nuc_single)
    return single_cell_img, nucleus_single_cell_img, single_cell_mask, df_nuc_single


def single_cell_analysis(
    single_cell_img,
    single_cell_mask,
    df_nuc_single,
    x_fiber,
    y_fiber,
    cell_label,
    internalised_threshold=0.75,
):
    n_nuc, n_nuc_intern, n_nuc_periph = 0, 0, 0
    new_row_lst = []
    new_col_names = df_nuc_single.columns.tolist()
    new_col_names.append("internalised")
    new_col_names.append("score_eccentricity")
    new_col_names.append("cell label")
    for _, value in df_nuc_single.iterrows():
        n_nuc += 1
        value = value.tolist()
        # Extend line and find closest point

        # Handling of the case where the nucleus is at the exact center of the fiber
        if x_fiber == value[3] and y_fiber == value[2]:
            n_nuc_intern += 1
            continue

        m, b = line_equation(x_fiber, y_fiber, value[3], value[2])

        intersections_lst = calculate_intersection(
            m, b, (single_cell_img.shape[0], single_cell_img.shape[1])
        )
        border_point = calculate_closest_point(value[3], value[2], intersections_lst)
        rr, cc = line(
            int(y_fiber),
            int(x_fiber),
            int(border_point[1]),
            int(border_point[0]),
        )
        for index3, coords in enumerate(list(zip(rr, cc))):
            try:
                if single_cell_mask[coords] == 0:
                    dist_nuc_cent = calculate_distance(
                        x_fiber, y_fiber, value[3], value[2]
                    )
                    dist_out_of_fiber = calculate_distance(
                        x_fiber, y_fiber, coords[1], coords[0]
                    )
                    ratio_dist = dist_nuc_cent / dist_out_of_fiber
                    if ratio_dist <= internalised_threshold:
                        n_nuc_intern += 1
                        value.append(True)
                        value.append(ratio_dist)
                        value.append(cell_label)
                    else:
                        n_nuc_periph += 1
                        value.append(False)
                        value.append(ratio_dist)
                        value.append(cell_label)
                    new_row_lst.append(value)
                    break
            except IndexError:
                coords = list(zip(rr, cc))[index3 - 1]
                dist_nuc_cent = calculate_distance(x_fiber, y_fiber, value[3], value[2])
                dist_out_of_fiber = calculate_distance(
                    x_fiber, y_fiber, coords[1], coords[0]
                )
                ratio_dist = dist_nuc_cent / dist_out_of_fiber
                if ratio_dist < internalised_threshold:
                    n_nuc_intern += 1
                    value.append(True)
                    value.append(ratio_dist)
                    value.append(cell_label)
                else:
                    n_nuc_periph += 1
                    value.append(True)
                    value.append(ratio_dist)
                    value.append(cell_label)
                new_row_lst.append(value)
                break
    df_nuc_single_stats = pd.DataFrame(new_row_lst, columns=new_col_names)
    return n_nuc, n_nuc_intern, n_nuc_periph, df_nuc_single_stats


def predict_all_cells(
    histo_img, cellpose_df, mask_stardist, internalised_threshold=0.75
):
    list_n_nuc, list_n_nuc_intern, list_n_nuc_periph, list_nuc_df = [], [], [], []
    for index in range(len(cellpose_df)):
        (
            single_cell_img,
            _,
            single_cell_mask,
            df_nuc_single,
        ) = extract_ROIs(histo_img, index, cellpose_df, mask_stardist)
        x_fiber = cellpose_df.iloc[index, 3] - cellpose_df.iloc[index, 6]
        y_fiber = cellpose_df.iloc[index, 2] - cellpose_df.iloc[index, 5]
        n_nuc, n_nuc_intern, n_nuc_periph, df_nuc_single_stats = single_cell_analysis(
            single_cell_img,
            single_cell_mask,
            df_nuc_single,
            x_fiber,
            y_fiber,
            index + 1,
            internalised_threshold,
        )
        list_n_nuc.append(n_nuc)
        list_n_nuc_intern.append(n_nuc_intern)
        list_n_nuc_periph.append(n_nuc_periph)
        list_nuc_df.append(df_nuc_single_stats)
        df_nuc_analysis = pd.DataFrame(
            list(zip(list_n_nuc, list_n_nuc_intern, list_n_nuc_periph)),
            columns=["N° Nuc", "N° Nuc Intern", "N° Nuc Periph"],
        )
    all_nuc_df_stats = pd.concat(list_nuc_df, ignore_index=True)
    cellpose_df_stat = pd.concat([cellpose_df, df_nuc_analysis], axis=1)
    return cellpose_df_stat, all_nuc_df_stats


def paint_histo_img(histo_img, cellpose_df, prediction_df):
    paint_img = np.zeros((histo_img.shape[0], histo_img.shape[1]), dtype=np.uint16)
    for index in range(len(cellpose_df)):
        single_cell_mask = cellpose_df.iloc[index, 9].copy()
        if prediction_df["N° Nuc Intern"].iloc[index] == 0:
            paint_img[
                cellpose_df.iloc[index, 5] : cellpose_df.iloc[index, 7],
                cellpose_df.iloc[index, 6] : cellpose_df.iloc[index, 8],
            ][single_cell_mask] = 1
        elif prediction_df["N° Nuc Intern"].iloc[index] > 0:
            paint_img[
                cellpose_df.iloc[index, 5] : cellpose_df.iloc[index, 7],
                cellpose_df.iloc[index, 6] : cellpose_df.iloc[index, 8],
            ][single_cell_mask] = 2
    return paint_img


def run_he_analysis(image_ndarray, mask_cellpose, mask_stardist, eccentricity_thresh):
    props_cellpose = regionprops_table(
        mask_cellpose,
        properties=[
            "label",
            "area",
            "centroid",
            "eccentricity",
            "bbox",
            "image",
            "perimeter",
            "feret_diameter_max",
        ],
    )
    df_cellpose = pd.DataFrame(props_cellpose)
    df_nuc_analysis, all_nuc_df_stats = predict_all_cells(
        image_ndarray, df_cellpose, mask_stardist, eccentricity_thresh
    )

    # Result table dict
    headers = ["Feature", "Raw Count", "Proportion (%)"]
    data = []
    data.append(["N° Nuclei", df_nuc_analysis["N° Nuc"].sum(), 100])
    data.append(
        [
            "N° Intern. Nuclei",
            df_nuc_analysis["N° Nuc Intern"].sum(),
            100
            * df_nuc_analysis["N° Nuc Intern"].sum()
            / df_nuc_analysis["N° Nuc"].sum(),
        ]
    )
    data.append(
        [
            "N° Periph. Nuclei",
            df_nuc_analysis["N° Nuc Periph"].sum(),
            100
            * df_nuc_analysis["N° Nuc Periph"].sum()
            / df_nuc_analysis["N° Nuc"].sum(),
        ]
    )
    data.append(
        [
            "N° Cells with 1+ intern. nuc.",
            df_nuc_analysis["N° Nuc Intern"].astype(bool).sum(axis=0),
            100
            * df_nuc_analysis["N° Nuc Intern"].astype(bool).sum(axis=0)
            / len(df_nuc_analysis),
        ]
    )

    result_df = pd.DataFrame(columns=headers, data=data)
    label_map_he = paint_histo_img(image_ndarray, df_cellpose, df_nuc_analysis)
    return result_df, label_map_he, df_nuc_analysis, all_nuc_df_stats



"""
CLI Command that automatically generates documentation for the CLI commands 
including options and args values. 
Source: https://github.com/tiangolo/typer-cli/pull/67#issuecomment-1271983950
"""

from pathlib import Path
from typing import cast

import typer
import typer.core
from click import Command, Group


app = typer.Typer()


def get_docs_for_click(
    *,
    obj: Command,
    ctx: typer.Context,
    indent: int = 0,
    name: str = "",
    call_prefix: str = "",
) -> str:
    docs = "#" * (1 + indent)
    command_name = name or obj.name
    if call_prefix:
        command_name = f"{call_prefix} {command_name}"
    title = f"`{command_name}`" if command_name else "CLI"
    docs += f" {title}\n\n"
    if obj.help:
        docs += f"{obj.help}\n\n"
    usage_pieces = obj.collect_usage_pieces(ctx)
    if usage_pieces:
        docs += "**Usage**:\n\n"
        docs += "```console\n"
        docs += "$ "
        if command_name:
            docs += f"{command_name} "
        docs += f"{' '.join(usage_pieces)}\n"
        docs += "```\n\n"
    args = []
    opts = []
    for param in obj.get_params(ctx):
        rv = param.get_help_record(ctx)
        if rv is not None:
            if param.param_type_name == "argument":
                args.append(rv)
            elif param.param_type_name == "option":
                opts.append(rv)
    if args:
        docs += "**Arguments**:\n\n"
        for arg_name, arg_help in args:
            docs += f"* `{arg_name}`"
            if arg_help:
                docs += f": {arg_help}"
            docs += "\n"
        docs += "\n"
    if opts:
        docs += "**Options**:\n\n"
        for opt_name, opt_help in opts:
            docs += f"* `{opt_name}`"
            if opt_help:
                docs += f": {opt_help}"
            docs += "\n"
        docs += "\n"
    if obj.epilog:
        docs += f"{obj.epilog}\n\n"
    if isinstance(obj, Group):
        group: Group = cast(Group, obj)
        commands = group.list_commands(ctx)
        if commands:
            docs += "**Commands**:\n\n"
            for command in commands:
                command_obj = group.get_command(ctx, command)
                assert command_obj
                docs += f"* `{command_obj.name}`"
                command_help = command_obj.get_short_help_str()
                if command_help:
                    docs += f": {command_help}"
                docs += "\n"
            docs += "\n"
        for command in commands:
            command_obj = group.get_command(ctx, command)
            assert command_obj
            use_prefix = ""
            if command_name:
                use_prefix += f"{command_name}"
            docs += get_docs_for_click(
                obj=command_obj, ctx=ctx, indent=indent + 1, call_prefix=use_prefix
            )
    return docs


@app.command(help="Generate markdown version of usage documentation")
def generate(
    ctx: typer.Context,
    name: str = typer.Option("", help="The name of the CLI program to use in docs."),
    output: Path = typer.Option(
        None,
        help="An output file to write docs to, like README.md.",
        file_okay=True,
        dir_okay=False,
    ),
) -> None:
    """
    Generate Markdown docs for a Typer app.
    """
    from __main__ import app as main_app

    click_obj = typer.main.get_command(main_app)
    docs = get_docs_for_click(obj=click_obj, ctx=ctx, name=name)
    clean_docs = f"{docs.strip()}\n"
    if output:
        output.write_text(clean_docs)
        typer.echo(f"Docs saved to: {output}")
    else:
        typer.echo(clean_docs)



import typer
from rich.console import Console

from .commands.docs import app as docs_app
from .commands import run_sdh, run_he, run_atp

console = Console()

app = typer.Typer(
    name="MyoQuant",
    add_completion=False,
    help="MyoQuant Analysis Command Line Interface",
    pretty_exceptions_show_locals=False,
)
app.add_typer(docs_app, name="docs", help="Generate documentation")


app.registered_commands += (
    run_sdh.app.registered_commands
    + run_he.app.registered_commands
    + run_atp.app.registered_commands
)

if __name__ == "__main__":
    app()



"""
Module that contains the main function to run the nuclei analysis for HE and fluo images
"""
import time
from pathlib import Path
import typer
from rich.console import Console
from rich.table import Table
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn

app = typer.Typer()
console = Console()
table = Table(title="Analysis Results 🥳")


@app.command()
def he_analysis(
    image_path: Path = typer.Argument(
        ...,
        help="The HE image file path to analyse. If using single channel images, this will be used as cytoplasm image to run CellPose. Please use the --fluo-nuc option to indicate the path to the nuclei single image to run Stardist.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    mask_path: Path = typer.Option(
        None,
        help="The path to a binary mask to hide slide region during analysis. It needs to be of the same resolution as input image and only pixel marked as 1 will be analyzed.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    cellpose_path: Path = typer.Option(
        None,
        help="The pre-computed CellPose mask to use for analysis. Will run Cellpose if no path provided. Required as an image file.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    stardist_path: Path = typer.Option(
        None,
        help="The pre-computed Stardist mask to use for analysis. Will run Stardist if no path provided. Required as an image file.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    output_path: Path = typer.Option(
        None,
        help="The path to the folder to save the results. Will save in the same folder as input image if not specified.",
    ),
    cellpose_diameter: int = typer.Option(
        None,
        help="Approximative single cell diameter in pixel for CellPose detection. If not specified, Cellpose will try to deduce it.",
    ),
    nms_thresh: float = typer.Option(
        0.4, help="NMS Threshold for Stardist nuclei detection.", min=0, max=1
    ),
    prob_thresh: float = typer.Option(
        0.5, help="Probability Threshold for Stardist nuclei detection.", min=0.5, max=1
    ),
    eccentricity_thresh: float = typer.Option(
        0.75,
        help="Eccentricity threshold value for a nucleus to be considered as internalized during nuclei classification. When very close to 1 almost all nuclei are considered as internalized.",
        min=0,
        max=1,
    ),
    export_map: bool = typer.Option(
        True,
        help="Export the original image with cells painted by classification label.",
    ),
    export_stats: bool = typer.Option(
        True, help="Export per fiber and per nuclei stat table."
    ),
    fluo_nuc: Path = typer.Option(
        None,
        help="The path to single channel fluo image for nuclei.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
):
    """Run the nuclei position analysis on HE and fluo images.
    First input arguments and option are printed in stdout and all modules are imported. Then the input image is mask with the binary mask if provided.
    Then depending on the presence of cellpose and stardist path, Cellpose and Stardist are run or not and mask accordingly if binary mask is provided.
    Finally the nuclei analysis is run with run_he_analysis() function and the results are saved in the output folder and some info are printed in stdout.
    """
    start_time = time.time()
    console.print(
        "👋 [bold dark_orange]Welcome to the nuclei position analysis (HE and fluo images)",
    )

    # Print input arguments and options
    console.print(f"📄 INPUT: raw image: {image_path}", style="blue")

    if fluo_nuc is not None:
        console.print(f"📄 INPUT: raw nuclear fluo image: {fluo_nuc}", style="blue")

    if cellpose_path is None:
        console.print(
            "💡 INFO: No CellPose mask provided, will run CellPose during the analysis.",
            style="blue",
        )
    else:
        console.print(f"📄 INPUT: CellPose mask: {cellpose_path}", style="blue")

    if stardist_path is None:
        console.print(
            "💡 INFO: No Stardist mask provided, will run Stardist during the analysis.",
            style="blue",
        )
    else:
        console.print(f"📄 INPUT: Stardist mask: {stardist_path}", style="blue")
    if mask_path is not None:
        console.print(f"📄 INPUT: binary mask: {mask_path}", style="blue")

    # Import all modules
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        transient=False,
    ) as progress:
        progress.add_task(description="Importing the libraries...", total=None)
        from ..src.common_func import (
            is_gpu_availiable,
            load_cellpose,
            load_stardist,
            run_cellpose,
            run_stardist,
            label2rgb,
            blend_image_with_label,
            HiddenPrints,
        )
        from ..src.HE_analysis import run_he_analysis
        import numpy as np
        from PIL import Image

        try:
            from imageio.v2 import imread
        except ImportError:
            from imageio import imread

    if output_path is None:
        output_path = image_path.parents[0]
    else:
        Path(output_path).mkdir(parents=True, exist_ok=True)

    if is_gpu_availiable():
        console.print("💡 INFO: GPU is available.", style="blue")
    else:
        console.print("❌ INFO: GPU is not available. Using CPU only.", style="red")

    # Load raw image, binary mask, cellpose and stardist mask if provided.
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        transient=False,
    ) as progress:
        progress.add_task(description="Reading all inputs...", total=None)
        image_ndarray = imread(image_path)

        if fluo_nuc is not None:
            fluo_nuc_ndarray = imread(fluo_nuc)
            mix_cyto_nuc = np.maximum(image_ndarray, fluo_nuc_ndarray)
        if mask_path is not None:
            mask_ndarray = imread(mask_path)
            if np.unique(mask_ndarray).shape[0] != 2:
                console.print(
                    "The mask image should be a binary image with only 2 values (0 and 1).",
                    style="red",
                )
                raise ValueError
            if len(image_ndarray.shape) > 2:
                mask_ndarray = np.repeat(
                    mask_ndarray.reshape(
                        mask_ndarray.shape[0], mask_ndarray.shape[1], 1
                    ),
                    image_ndarray.shape[2],
                    axis=2,
                )
            image_ndarray = image_ndarray * mask_ndarray
            if fluo_nuc is not None:
                fluo_nuc_ndarray = fluo_nuc_ndarray * mask_ndarray
        if cellpose_path is not None:
            mask_cellpose = imread(cellpose_path)
        if stardist_path is not None:
            mask_stardist = imread(stardist_path)

    # Run Cellpose if no mask provided
    if cellpose_path is None:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            transient=False,
        ) as progress:
            progress.add_task(description="Running CellPose...", total=None)
            model_cellpose = load_cellpose()
            mask_cellpose = run_cellpose(
                image_ndarray, model_cellpose, cellpose_diameter
            )
            mask_cellpose = mask_cellpose.astype(np.uint16)
            cellpose_mask_filename = image_path.stem + "_cellpose_mask.tiff"
            Image.fromarray(mask_cellpose).save(output_path / cellpose_mask_filename)
        console.print(
            f"💾 OUTPUT: CellPose mask saved as {output_path/cellpose_mask_filename}",
            style="green",
        )

    # Run Stardist if no mask provided
    if stardist_path is None:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            transient=False,
        ) as progress:
            progress.add_task(description="Running Stardist...", total=None)
            with HiddenPrints():
                if fluo_nuc is None:
                    model_stardist = load_stardist(fluo=False)
                else:
                    model_stardist = load_stardist(fluo=True)
            if fluo_nuc is not None:
                with HiddenPrints():
                    mask_stardist = run_stardist(
                        fluo_nuc_ndarray, model_stardist, nms_thresh, prob_thresh
                    )
            else:
                with HiddenPrints():
                    mask_stardist = run_stardist(
                        image_ndarray, model_stardist, nms_thresh, prob_thresh
                    )
            mask_stardist = mask_stardist.astype(np.uint16)
            stardist_mask_filename = image_path.stem + "_stardist_mask.tiff"
            Image.fromarray(mask_stardist).save(output_path / stardist_mask_filename)
        console.print(
            f"💾 OUTPUT: Stardist mask saved as {output_path/stardist_mask_filename}",
            style="green",
        )

    # If binary mask provided, mask cellpose and stardist mask
    if mask_path is not None:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            transient=False,
        ) as progress:
            progress.add_task(
                description="Masking Cellpose and Stardist mask with binary mask...",
                total=None,
            )
            mask_ndarray = imread(mask_path)
            mask_stardist = mask_stardist * mask_ndarray
            mask_cellpose = mask_cellpose * mask_ndarray

    # Run the nuclei position analysis and get the results table, label map and dataframes
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        transient=False,
    ) as progress:
        progress.add_task(
            description="Calculating all nuclei eccentricity scores...", total=None
        )
        result_df, full_label_map, df_nuc_analysis, all_nuc_df_stats = run_he_analysis(
            image_ndarray, mask_cellpose, mask_stardist, eccentricity_thresh
        )
    if export_map:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            transient=False,
        ) as progress:
            progress.add_task(
                description="Blending label and original image together...", total=None
            )
            if fluo_nuc is not None:
                labelRGB_map = label2rgb(mix_cyto_nuc, full_label_map)
                overlay_img = blend_image_with_label(
                    mix_cyto_nuc, labelRGB_map, fluo=True
                )
            else:
                labelRGB_map = label2rgb(image_ndarray, full_label_map)
                overlay_img = blend_image_with_label(image_ndarray, labelRGB_map)
            overlay_filename = image_path.stem + "_label_blend.tiff"
            overlay_img.save(output_path / overlay_filename)

    # Construct the summary table, print all output in stdout and save files in output folder.
    table.add_column("Feature", justify="left", style="cyan")
    table.add_column("Raw Count", justify="center", style="magenta")
    table.add_column("Proportion (%)", justify="right", style="green")
    for index, row in result_df.iterrows():
        table.add_row(
            str(row[0]),
            str(row[1]),
            str(row[2]),
        )
    console.print(table)
    csv_name = image_path.stem + "_results_summary.csv"
    cell_details_name = image_path.stem + "_cell_details.csv"
    nuc_details_name = image_path.stem + "_nuc_details.csv"
    result_df.to_csv(
        output_path / csv_name,
        index=False,
    )
    console.print(
        f"💾 OUTPUT: Summary Table saved as {output_path/csv_name}",
        style="green",
    )
    if export_map:
        console.print(
            f"💾 OUTPUT: Overlay image saved as {output_path/overlay_filename}",
            style="green",
        )
    if export_stats:
        df_nuc_analysis.drop("image", axis=1).to_csv(
            output_path / cell_details_name,
            index=False,
        )
        console.print(
            f"💾 OUTPUT: Cell Table saved as {output_path/cell_details_name}",
            style="green",
        )
        all_nuc_df_stats.drop("image", axis=1).to_csv(
            output_path / nuc_details_name,
            index=False,
        )
        console.print(
            f"💾 OUTPUT: Nuclei Table saved as {output_path/nuc_details_name}",
            style="green",
        )
    label_map_name = image_path.stem + "_label_map.tiff"
    Image.fromarray(full_label_map).save(output_path / label_map_name)
    console.print(
        f"💾 OUTPUT: Label map saved as {output_path/label_map_name}", style="green"
    )
    console.print("--- %s seconds ---" % (int(time.time() - start_time)))



"""
Module that contains the main function to run the mitochondrial distribution analysis for SDH images
"""
import time
from pathlib import Path
import typer
from rich.console import Console
from rich.table import Table
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn

app = typer.Typer()
console = Console()
table = Table(title="Analysis Results 🥳")


@app.command()
def sdh_analysis(
    image_path: Path = typer.Argument(
        ...,
        help="The image file path to analyse.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    mask_path: Path = typer.Option(
        None,
        help="The path to a binary mask to hide slide region during analysis. It needs to be of the same resolution as input image and only pixel marked as 1 will be analyzed.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    model_path: Path = typer.Option(
        None,
        help="The SDH model path to use for analysis. Will download latest one if no path provided.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    cellpose_path: Path = typer.Option(
        None,
        help="The pre-computed CellPose mask to use for analysis. Will run Cellpose if no path provided. Required as an image file.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    output_path: Path = typer.Option(
        None,
        help="The path to the folder to save the results. Will save in the current folder if not specified.",
    ),
    cellpose_diameter: int = typer.Option(
        None,
        help="Approximative single cell diameter in pixel for CellPose detection. If not specified, Cellpose will try to deduce it.",
    ),
    export_map: bool = typer.Option(
        True,
        help="Export the original image with cells painted by classification label.",
    ),
    export_stats: bool = typer.Option(True, help="Export per fiber stat table."),
):
    """Run the mitochondiral analysis and quantification on the image.
    First input arguments and option are printed in stdout and all modules are imported and latest SDH model is downloaded.
    Then the input image is mask with the binary mask if provided.
    Then depending on the presence of cellpose path, Cellpose is run or not and mask accordingly if binary mask is provided.
    Finally the mitochondiral classificaiton is run with run_sdh_analysis() function and the results are saved in the output folder and some info are printed in stdout.
    """
    start_time = time.time()
    console.print(
        "👋 [bold dark_orange]Welcome to the mitochondrial distribution analysis (SDH images)",
    )
    # Print input arguments and options
    console.print(f"📄 INPUT: raw image: {image_path}", style="blue")

    if cellpose_path is None:
        console.print(
            "💡 INFO: No CellPose mask provided, will run CellPose during the analysis.",
            style="blue",
        )
    else:
        console.print(f"📄 INPUT: CellPose mask: {cellpose_path}", style="blue")

    if mask_path is not None:
        console.print(f"📄 INPUT: binary mask: {mask_path}", style="blue")
    import os

    # If the model path is not provided, download latest version or check existence.
    if model_path is None:
        import urllib.request

        console.print(
            "💡 INFO: No SDH model provided, will download or use latest one.",
            style="blue",
        )
        model_path_abs = Path(os.path.abspath(__file__)).parents[0] / "model.h5"
        if not os.path.exists(model_path_abs):
            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                TimeElapsedColumn(),
                transient=False,
            ) as progress:
                progress.add_task(description="Downloading SDH Model...", total=None)
                urllib.request.urlretrieve(
                    "https://lbgi.fr/~meyer/SDH_models/model.h5",
                    model_path_abs,
                )
        model_path = model_path_abs
    console.print(f"📄 INPUT: SDH Model: {model_path}", style="blue")

    # Import all modules
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        transient=False,
    ) as progress:
        progress.add_task(description="Importing the libraries...", total=None)
        from ..src.common_func import (
            is_gpu_availiable,
            load_cellpose,
            load_sdh_model,
            run_cellpose,
            label2rgb,
            blend_image_with_label,
            HiddenPrints,
        )
        from ..src.SDH_analysis import run_sdh_analysis
        import numpy as np
        from PIL import Image

        try:
            from imageio.v2 import imread
        except ImportError:
            from imageio import imread

    if output_path is None:
        output_path = image_path.parents[0]
    else:
        Path(output_path).mkdir(parents=True, exist_ok=True)

    if is_gpu_availiable():
        console.print("💡 INFO: GPU is available.", style="blue")
    else:
        console.print("❌ INFO: GPU is not available. Using CPU only.", style="red")

    # Load raw image, binary mask, cellpose and stardist mask if provided.
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        transient=False,
    ) as progress:
        progress.add_task(description="Reading all inputs...", total=None)
        image_ndarray_sdh = imread(image_path)

    if mask_path is not None:
        mask_ndarray = imread(mask_path)
        if np.unique(mask_ndarray).shape[0] != 2:
            console.print(
                "The mask image should be a binary image with only 2 values (0 and 1).",
                style="red",
            )
            raise ValueError
        if len(image_ndarray_sdh.shape) > 2:
            mask_ndarray = np.repeat(
                mask_ndarray.reshape(mask_ndarray.shape[0], mask_ndarray.shape[1], 1),
                image_ndarray_sdh.shape[2],
                axis=2,
            )
        image_ndarray_sdh = image_ndarray_sdh * mask_ndarray
        if cellpose_path is not None:
            mask_cellpose = imread(cellpose_path)

    # Run Cellpose if no mask provided
    if cellpose_path is None:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            transient=False,
        ) as progress:
            progress.add_task(description="Running CellPose...", total=None)
            model_cellpose = load_cellpose()
            mask_cellpose = run_cellpose(
                image_ndarray_sdh, model_cellpose, cellpose_diameter
            )
            mask_cellpose = mask_cellpose.astype(np.uint16)
            cellpose_mask_filename = image_path.stem + "_cellpose_mask.tiff"
            Image.fromarray(mask_cellpose).save(output_path / cellpose_mask_filename)
        console.print(
            f"💾 OUTPUT: CellPose mask saved as {output_path/cellpose_mask_filename}",
            style="green",
        )

    # Load Tensorflow SDH Model
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        transient=False,
    ) as progress:
        progress.add_task(
            description="Loading SDH Model...",
            total=None,
        )
        with HiddenPrints():
            model_SDH = load_sdh_model(model_path)

    # If binary mask provided, mask cellpose mask
    if mask_path is not None:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            transient=False,
        ) as progress:
            progress.add_task(
                description="Masking Cellpose and Stardist mask with binary mask...",
                total=None,
            )
        mask_ndarray = imread(mask_path)
        mask_cellpose = mask_cellpose * mask_ndarray

    # Run the mitoC distribution analysis and get the results table, label map and dataframes
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        transient=False,
    ) as progress:
        progress.add_task(
            description="Predicting all muscle fibers class...", total=None
        )
        with HiddenPrints():
            result_df, full_label_map, df_cellpose_details = run_sdh_analysis(
                image_ndarray_sdh, model_SDH, mask_cellpose
            )
    if export_map:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            transient=False,
        ) as progress:
            progress.add_task(
                description="Blending label and original image together...", total=None
            )
            labelRGB_map = label2rgb(image_ndarray_sdh, full_label_map)
            overlay_img = blend_image_with_label(image_ndarray_sdh, labelRGB_map)
            overlay_filename = image_path.stem + "_label_blend.tiff"
            overlay_img.save(output_path / overlay_filename)

    # Construct the summary table, print all output in stdout and save files in output folder.
    table.add_column("Feature", justify="left", style="cyan")
    table.add_column("Raw Count", justify="center", style="magenta")
    table.add_column("Proportion (%)", justify="right", style="green")
    for index, row in result_df.iterrows():
        table.add_row(
            str(row[0]),
            str(row[1]),
            str(row[2]),
        )
    console.print(table)
    csv_name = image_path.stem + "_results.csv"
    result_df.to_csv(
        output_path / csv_name,
        index=False,
    )
    console.print(
        f"💾 OUTPUT: Summary Table saved as {output_path/csv_name}", style="green"
    )
    if export_stats:
        cell_details_name = image_path.stem + "_cell_details.csv"
        df_cellpose_details.drop("image", axis=1).to_csv(
            output_path / cell_details_name,
            index=False,
        )
        console.print(
            f"💾 OUTPUT: Cell Table saved as {output_path/cell_details_name}",
            style="green",
        )
    label_map_name = image_path.stem + "_label_map.tiff"
    Image.fromarray(full_label_map).save(output_path / label_map_name)
    console.print(
        f"💾 OUTPUT: Label map saved as {output_path/label_map_name}", style="green"
    )
    console.print("--- %s seconds ---" % int((time.time() - start_time)))



import pandas as pd
from skimage.measure import regionprops_table
from scipy.stats import gaussian_kde
from sklearn.mixture import GaussianMixture

import numpy as np

labels_predict = {1: "fiber type 1", 2: "fiber type 2"}
np.random.seed(42)


def get_all_intensity(image_array, df_cellpose):
    all_cell_median_intensity = []
    for index in range(len(df_cellpose)):
        single_cell_img = image_array[
            df_cellpose.iloc[index, 5] : df_cellpose.iloc[index, 7],
            df_cellpose.iloc[index, 6] : df_cellpose.iloc[index, 8],
        ].copy()

        single_cell_mask = df_cellpose.iloc[index, 9].copy()
        single_cell_img[~single_cell_mask] = 0
        # Calculate median pixel intensity of the cell but ignore 0 values
        single_cell_median_intensity = np.median(single_cell_img[single_cell_img > 0])
        all_cell_median_intensity.append(single_cell_median_intensity)
    return all_cell_median_intensity


def estimate_threshold(intensity_list):
    density = gaussian_kde(intensity_list)
    density.covariance_factor = lambda: 0.25
    density._compute_covariance()

    # Create a vector of 256 values going from 0 to 256:
    xs = np.linspace(0, 255, 256)
    density_xs_values = density(xs)
    gmm = GaussianMixture(n_components=2).fit(np.array(intensity_list).reshape(-1, 1))

    # Find the x values of the two peaks
    peaks_x = np.sort(gmm.means_.flatten())
    # Find the minimum point between the two peaks
    min_index = np.argmin(density_xs_values[(xs > peaks_x[0]) & (xs < peaks_x[1])])
    threshold = peaks_x[0] + xs[min_index]

    return threshold


def predict_all_cells(histo_img, cellpose_df, intensity_threshold):
    all_cell_median_intensity = get_all_intensity(histo_img, cellpose_df)
    if intensity_threshold is None:
        intensity_threshold = estimate_threshold(all_cell_median_intensity)

    muscle_fiber_type_all = [
        1 if x > intensity_threshold else 2 for x in all_cell_median_intensity
    ]
    return muscle_fiber_type_all, all_cell_median_intensity


def paint_full_image(image_atp, df_cellpose, class_predicted_all):
    image_atp_paint = np.zeros(
        (image_atp.shape[0], image_atp.shape[1]), dtype=np.uint16
    )
    # for index in track(range(len(df_cellpose)), description="Painting cells"):
    for index in range(len(df_cellpose)):
        single_cell_mask = df_cellpose.iloc[index, 9].copy()
        if class_predicted_all[index] == 1:
            image_atp_paint[
                df_cellpose.iloc[index, 5] : df_cellpose.iloc[index, 7],
                df_cellpose.iloc[index, 6] : df_cellpose.iloc[index, 8],
            ][single_cell_mask] = 1
        elif class_predicted_all[index] == 2:
            image_atp_paint[
                df_cellpose.iloc[index, 5] : df_cellpose.iloc[index, 7],
                df_cellpose.iloc[index, 6] : df_cellpose.iloc[index, 8],
            ][single_cell_mask] = 2
    return image_atp_paint


def run_atp_analysis(image_array, mask_cellpose, intensity_threshold=None):
    props_cellpose = regionprops_table(
        mask_cellpose,
        properties=[
            "label",
            "area",
            "centroid",
            "eccentricity",
            "bbox",
            "image",
            "perimeter",
        ],
    )
    df_cellpose = pd.DataFrame(props_cellpose)
    class_predicted_all, intensity_all = predict_all_cells(
        image_array, df_cellpose, intensity_threshold
    )
    df_cellpose["muscle_cell_type"] = class_predicted_all
    df_cellpose["cell_intensity"] = intensity_all
    count_per_label = np.unique(class_predicted_all, return_counts=True)

    # Result table dict
    headers = ["Feature", "Raw Count", "Proportion (%)"]
    data = []
    data.append(["Muscle Fibers", len(class_predicted_all), 100])
    for index, elem in enumerate(count_per_label[0]):
        data.append(
            [
                labels_predict[int(elem)],
                count_per_label[1][int(index)],
                100 * count_per_label[1][int(index)] / len(class_predicted_all),
            ]
        )
    result_df = pd.DataFrame(columns=headers, data=data)
    # Paint The Full Image
    full_label_map = paint_full_image(image_array, df_cellpose, class_predicted_all)
    return result_df, full_label_map, df_cellpose
